@@ -6,38 +6,45 @@
 #include "mpi.h"
 
 
-#define OBJ2C(rb_obj, len, buffer, typ) \
+#define OBJ2C(rb_obj, len, buffer, typ, off) \
 {\
   if (TYPE(rb_obj) == T_STRING) {\
-    len = RSTRING_LEN(rb_obj);\
-    buffer = (void*)StringValuePtr(rb_obj);\
-    typ = MPI_CHAR;\
-  } else if (IsNArray(rb_obj)) {		\
+    if (len==0) len = RSTRING_LEN(rb_obj);\
+    buffer = (void*)(StringValuePtr(rb_obj) + off);\
+    typ = MPI_BYTE;\
+  } else if (IsNArray(rb_obj)) {\
     struct NARRAY *a;\
     GetNArray(rb_obj, a);\
     buffer = (void*)(a->ptr);\
-    len = a->total;\
+    if (len==0) len = a->total;\
     switch (a->type) {\
     case NA_BYTE:\
       typ = MPI_BYTE;\
+      buffer = (void*)((char*)buffer + off);\
       break;\
     case NA_SINT:\
       typ = MPI_SHORT;\
+      buffer = (void*)((char*)buffer + off*4);\
       break;\
     case NA_LINT:\
       typ = MPI_LONG;\
+      buffer = (void*)((char*)buffer + off*8);\
       break;\
     case NA_SFLOAT:\
       typ = MPI_FLOAT;\
+      buffer = (void*)((char*)buffer + off*4);\
       break;\
     case NA_DFLOAT:\
       typ = MPI_DOUBLE;\
+      buffer = (void*)((char*)buffer + off*8);\
       break;\
     case NA_SCOMPLEX:\
       typ = MPI_2COMPLEX;\
+      buffer = (void*)((char*)buffer + off*8);\
       break;\
     case NA_DCOMPLEX:\
       typ = MPI_2DOUBLE_COMPLEX;\
+      buffer = (void*)((char*)buffer + off*16);\
       break;\
     default:\
       rb_raise(rb_eArgError, "narray type is invalid");\
@@ -54,48 +61,36 @@ static VALUE eBUFFER, eCOUNT, eTYPE, eTAG, eCOMM, eRANK, eREQUEST, eROOT, eGROUP
 
 struct _Comm {
   MPI_Comm Comm;
+  bool free;
 };
 struct _Request {
   MPI_Request Request;
+  bool free;
 };
 struct _Op {
   MPI_Op Op;
+  bool free;
 };
 struct _Errhandler {
   MPI_Errhandler Errhandler;
+  bool free;
 };
 
 static bool _initialized = false;
 static bool _finalized = false;
 
 
-#define DEF_FREE(name) \
-static void \
-name ## _free(void *ptr)\
-{\
-  struct _ ## name *obj;\
-  obj = (struct _ ## name*) ptr;\
-  if (!_finalized)\
-    MPI_ ## name ## _free(&(obj->name)); \
-  free(obj);\
-}
-DEF_FREE(Comm)
-DEF_FREE(Request)
-DEF_FREE(Op)
-DEF_FREE(Errhandler)
-static void
-Status_free(void *ptr)
-{
-  free((MPI_Status*) ptr);
-}
-
-
-#define CAE_ERR(type) case MPI_ERR_ ## type: rb_raise(e ## type,""); break
+#define CAE_ERR(type) case MPI_ERR_ ## type: rb_raise(e ## type,"%s",str); break
 static void
 check_error(int error)
 {
-  switch (error) {
-  case MPI_SUCCESS: break;
+  if (error == MPI_SUCCESS) return;
+  int code, len;
+  char str[MPI_MAX_ERROR_STRING];
+  if (MPI_Error_class(error, &code)!=MPI_SUCCESS || MPI_Error_string(error, str, &len)!=MPI_SUCCESS)
+    rb_raise(rb_eRuntimeError, "unknown error occuerd in MPI call");
+
+  switch (code) {
     CAE_ERR(BUFFER);
     CAE_ERR(COUNT);
     CAE_ERR(TYPE);
@@ -154,15 +149,48 @@ check_error(int error)
     CAE_ERR(SYSRESOURCE);
 #endif
   default:
-    rb_raise(rb_eRuntimeError, "unknown error");
+    rb_raise(rb_eRuntimeError, "unknown error: %d", code);
   }
 }
 
+#define DEF_FREE(name, capit)				\
+static void \
+name ## _free(void *ptr)\
+{\
+  struct _ ## name *obj;\
+  obj = (struct _ ## name*) ptr;\
+  if (!_finalized && obj->free && obj->name!=MPI_ ## capit ##_NULL)\
+    check_error(MPI_ ## name ## _free(&(obj->name))); \
+  free(obj);\
+}
+#define DEF_FREE2(name, capit)				\
+static void \
+name ## _free2(void *ptr)\
+{\
+  struct _ ## name *obj;\
+  obj = (struct _ ## name*) ptr;\
+  free(obj);\
+}
+DEF_FREE(Comm, COMM)
+DEF_FREE(Request, REQUEST)
+DEF_FREE(Op, OP)
+DEF_FREE(Errhandler, ERRHANDLER)
+DEF_FREE2(Comm, COMM)
+DEF_FREE2(Op, OP)
+DEF_FREE2(Errhandler, ERRHANDLER)
+static void
+Status_free(void *ptr)
+{
+  free((MPI_Status*) ptr);
+}
+
+
 #define DEF_CONST(v, const, name) \
 {\
   v = ALLOC(struct _ ## v);\
   v->v = const;\
-  rb_define_const(c ## v, #name, Data_Wrap_Struct(c ## v, NULL, v ## _free, v));	\
+  v->free = false;\
+  rb_define_const(c ## v, #name, Data_Wrap_Struct(c ## v, NULL, v ## _free2, v));	\
 }
 
 static void
@@ -216,7 +244,7 @@ rb_m_init(int argc, VALUE *argv, VALUE self)
   // define MPI::Comm::WORLD
   struct _Comm *Comm;
   DEF_CONST(Comm, MPI_COMM_WORLD, WORLD);
-  MPI_Errhandler_set(MPI_COMM_WORLD, MPI_ERRORS_RETURN);
+  check_error(MPI_Errhandler_set(MPI_COMM_WORLD, MPI_ERRORS_RETURN));
 
   // define MPI::Op::???
   struct _Op *Op;
@@ -249,6 +277,16 @@ rb_m_finalize(VALUE self)
   return self;
 }
 
+static VALUE
+rb_m_abort(VALUE self, VALUE rcomm, VALUE rerror)
+{
+  struct _Comm *comm;
+  int ierror;
+  Data_Get_Struct(rcomm, struct _Comm, comm);
+  ierror = MPI_Abort(comm->Comm, NUM2INT(rerror));
+  return INT2NUM(ierror);
+}
+
 
 // MPI::Comm
 static VALUE
@@ -262,6 +300,7 @@ rb_comm_initialize(VALUE self)
 {
   rb_raise(rb_eRuntimeError, "not developed yet");
   // MPI_Comm_create()
+  // comm->free = true;
 }
 static VALUE
 rb_comm_size(VALUE self)
@@ -285,11 +324,11 @@ static VALUE
 rb_comm_send(VALUE self, VALUE rb_obj, VALUE rb_dest, VALUE rb_tag)
 {
   void* buffer;
-  int len, dest, tag;
+  int len=0, dest, tag;
   MPI_Datatype type;
   struct _Comm *comm;
 
-  OBJ2C(rb_obj, len, buffer, type);
+  OBJ2C(rb_obj, len, buffer, type, 0);
   dest = NUM2INT(rb_dest);
   tag = NUM2INT(rb_tag);
   Data_Get_Struct(self, struct _Comm, comm);
@@ -301,31 +340,43 @@ static VALUE
 rb_comm_isend(VALUE self, VALUE rb_obj, VALUE rb_dest, VALUE rb_tag)
 {
   void* buffer;
-  int len, dest, tag;
+  int len=0, dest, tag;
   MPI_Datatype type;
   struct _Comm *comm;
   struct _Request *request;
   VALUE rb_request;
 
-  OBJ2C(rb_obj, len, buffer, type);
+  OBJ2C(rb_obj, len, buffer, type, 0);
   dest = NUM2INT(rb_dest);
   tag = NUM2INT(rb_tag);
   Data_Get_Struct(self, struct _Comm, comm);
   rb_request = Data_Make_Struct(cRequest, struct _Request, NULL, Request_free, request);
+  request->free = true;
   check_error(MPI_Isend(buffer, len, type, dest, tag, comm->Comm, &(request->Request)));
 
   return rb_request;
 }
 static VALUE
-rb_comm_recv(VALUE self, VALUE rb_obj, VALUE rb_source, VALUE rb_tag)
+rb_comm_recv(int argc, VALUE *argv, VALUE self)
 {
+  VALUE rb_obj, rb_source, rb_tag;
+  VALUE rb_len, rb_offset; // option
   void* buffer;
-  int len, source, tag;
+  int source, tag, len = 0, offset = 0;
   MPI_Datatype type;
   MPI_Status *status;
   struct _Comm *comm;
 
-  OBJ2C(rb_obj, len, buffer, type);
+  rb_scan_args(argc, argv, "32", &rb_obj, &rb_source, &rb_tag, &rb_len, &rb_offset);
+
+  if (rb_len != Qnil) {
+    len = NUM2INT(rb_len);
+  }
+  if (rb_offset != Qnil) {
+    offset = NUM2INT(rb_offset);
+  }
+
+  OBJ2C(rb_obj, len, buffer, type, offset);
   source = NUM2INT(rb_source);
   tag = NUM2INT(rb_tag);
 
@@ -336,20 +387,33 @@ rb_comm_recv(VALUE self, VALUE rb_obj, VALUE rb_source, VALUE rb_tag)
   return Data_Wrap_Struct(cStatus, NULL, Status_free, status);
 }
 static VALUE
-rb_comm_irecv(VALUE self, VALUE rb_obj, VALUE rb_source, VALUE rb_tag)
+rb_comm_irecv(int argc, VALUE *argv, VALUE self)
 {
+  VALUE rb_obj, rb_source, rb_tag;
+  VALUE rb_len, rb_offset; // option
   void* buffer;
-  int len, source, tag;
+  int source, tag, len = 0, offset = 0;
   MPI_Datatype type;
   struct _Comm *comm;
   struct _Request *request;
   VALUE rb_request;
 
-  OBJ2C(rb_obj, len, buffer, type);
+  rb_scan_args(argc, argv, "32", &rb_obj, &rb_source, &rb_tag, &rb_len, &rb_offset);
+
+  if (rb_len != Qnil) {
+    len = NUM2INT(rb_len);
+  }
+  if (rb_offset != Qnil) {
+    offset = NUM2INT(rb_offset);
+  }
+
+  OBJ2C(rb_obj, len, buffer, type, offset);
   source = NUM2INT(rb_source);
   tag = NUM2INT(rb_tag);
+
   Data_Get_Struct(self, struct _Comm, comm);
   rb_request = Data_Make_Struct(cRequest, struct _Request, NULL, Request_free, request);
+  request->free = true;
   check_error(MPI_Irecv(buffer, len, type, source, tag, comm->Comm, &(request->Request)));
 
   return rb_request;
@@ -358,17 +422,17 @@ static VALUE
 rb_comm_gather(VALUE self, VALUE rb_sendbuf, VALUE rb_recvbuf, VALUE rb_root)
 {
   void *sendbuf, *recvbuf = NULL;
-  int sendcount, recvcount = 0;
-  MPI_Datatype sendtype, recvtype = NULL;
+  int sendcount=0, recvcount = 0;
+  MPI_Datatype sendtype, recvtype = 0;
   int root, rank, size;
   struct _Comm *comm;
-  OBJ2C(rb_sendbuf, sendcount, sendbuf, sendtype);
+  OBJ2C(rb_sendbuf, sendcount, sendbuf, sendtype, 0);
   root = NUM2INT(rb_root);
   Data_Get_Struct(self, struct _Comm, comm);
   check_error(MPI_Comm_rank(comm->Comm, &rank));
   check_error(MPI_Comm_size(comm->Comm, &size));
   if (rank == root) {
-    OBJ2C(rb_recvbuf, recvcount, recvbuf, recvtype);
+    OBJ2C(rb_recvbuf, recvcount, recvbuf, recvtype, 0);
     if (recvcount < sendcount*size)
       rb_raise(rb_eArgError, "recvbuf is too small");
     recvcount = sendcount;
@@ -380,15 +444,15 @@ static VALUE
 rb_comm_allgather(VALUE self, VALUE rb_sendbuf, VALUE rb_recvbuf)
 {
   void *sendbuf, *recvbuf;
-  int sendcount, recvcount;
+  int sendcount=0, recvcount=0;
   MPI_Datatype sendtype, recvtype;
   int rank, size;
   struct _Comm *comm;
-  OBJ2C(rb_sendbuf, sendcount, sendbuf, sendtype);
+  OBJ2C(rb_sendbuf, sendcount, sendbuf, sendtype, 0);
   Data_Get_Struct(self, struct _Comm, comm);
   check_error(MPI_Comm_rank(comm->Comm, &rank));
   check_error(MPI_Comm_size(comm->Comm, &size));
-  OBJ2C(rb_recvbuf, recvcount, recvbuf, recvtype);
+  OBJ2C(rb_recvbuf, recvcount, recvbuf, recvtype, 0);
   if (recvcount < sendcount*size)
     rb_raise(rb_eArgError, "recvbuf is too small");
   recvcount = sendcount;
@@ -399,11 +463,11 @@ static VALUE
 rb_comm_bcast(VALUE self, VALUE rb_buffer, VALUE rb_root)
 {
   void *buffer;
-  int count;
+  int count=0;
   MPI_Datatype type;
   int root;
   struct _Comm *comm;
-  OBJ2C(rb_buffer, count, buffer, type);
+  OBJ2C(rb_buffer, count, buffer, type, 0);
   root = NUM2INT(rb_root);
   Data_Get_Struct(self, struct _Comm, comm);
   check_error(MPI_Bcast(buffer, count, type, root, comm->Comm));
@@ -413,17 +477,17 @@ static VALUE
 rb_comm_scatter(VALUE self, VALUE rb_sendbuf, VALUE rb_recvbuf, VALUE rb_root)
 {
   void *sendbuf = NULL, *recvbuf;
-  int sendcount = 0, recvcount;
-  MPI_Datatype sendtype = NULL, recvtype;
+  int sendcount = 0, recvcount=0;
+  MPI_Datatype sendtype = 0, recvtype;
   int root, rank, size;
   struct _Comm *comm;
-  OBJ2C(rb_recvbuf, recvcount, recvbuf, recvtype);
+  OBJ2C(rb_recvbuf, recvcount, recvbuf, recvtype, 0);
   root = NUM2INT(rb_root);
   Data_Get_Struct(self, struct _Comm, comm);
   check_error(MPI_Comm_rank(comm->Comm, &rank));
   check_error(MPI_Comm_size(comm->Comm, &size));
   if (rank == root) {
-    OBJ2C(rb_sendbuf, sendcount, sendbuf, sendtype);
+    OBJ2C(rb_sendbuf, sendcount, sendbuf, sendtype, 0);
     if (sendcount > recvcount*size)
       rb_raise(rb_eArgError, "recvbuf is too small");
     sendcount = recvcount;
@@ -435,17 +499,17 @@ static VALUE
 rb_comm_sendrecv(VALUE self, VALUE rb_sendbuf, VALUE rb_dest, VALUE rb_sendtag, VALUE rb_recvbuf, VALUE rb_source, VALUE rb_recvtag)
 {
   void *sendbuf, *recvbuf;
-  int sendcount, recvcount;
+  int sendcount=0, recvcount=0;
   MPI_Datatype sendtype, recvtype;
   int dest, source;
   int sendtag, recvtag;
   int size;
   struct _Comm *comm;
   MPI_Status *status;
-  OBJ2C(rb_sendbuf, sendcount, sendbuf, sendtype);
+  OBJ2C(rb_sendbuf, sendcount, sendbuf, sendtype, 0);
   Data_Get_Struct(self, struct _Comm, comm);
   check_error(MPI_Comm_size(comm->Comm, &size));
-  OBJ2C(rb_recvbuf, recvcount, recvbuf, recvtype);
+  OBJ2C(rb_recvbuf, recvcount, recvbuf, recvtype, 0);
   dest = NUM2INT(rb_dest);
   source = NUM2INT(rb_source);
   sendtag = NUM2INT(rb_sendtag);
@@ -458,14 +522,14 @@ static VALUE
 rb_comm_alltoall(VALUE self, VALUE rb_sendbuf, VALUE rb_recvbuf)
 {
   void *sendbuf, *recvbuf;
-  int sendcount, recvcount;
+  int sendcount=0, recvcount=0;
   MPI_Datatype sendtype, recvtype;
   int size;
   struct _Comm *comm;
-  OBJ2C(rb_sendbuf, sendcount, sendbuf, sendtype);
+  OBJ2C(rb_sendbuf, sendcount, sendbuf, sendtype, 0);
   Data_Get_Struct(self, struct _Comm, comm);
   check_error(MPI_Comm_size(comm->Comm, &size));
-  OBJ2C(rb_recvbuf, recvcount, recvbuf, recvtype);
+  OBJ2C(rb_recvbuf, recvcount, recvbuf, recvtype, 0);
   if (recvcount < sendcount)
     rb_raise(rb_eArgError, "recvbuf is too small");
   recvcount = recvcount/size;
@@ -477,18 +541,18 @@ static VALUE
 rb_comm_reduce(VALUE self, VALUE rb_sendbuf, VALUE rb_recvbuf, VALUE rb_op, VALUE rb_root)
 {
   void *sendbuf, *recvbuf = NULL;
-  int sendcount, recvcount = 0;
-  MPI_Datatype sendtype, recvtype = NULL;
+  int sendcount=0, recvcount = 0;
+  MPI_Datatype sendtype, recvtype = 0;
   int root, rank, size;
   struct _Comm *comm;
   struct _Op *op;
-  OBJ2C(rb_sendbuf, sendcount, sendbuf, sendtype);
+  OBJ2C(rb_sendbuf, sendcount, sendbuf, sendtype, 0);
   root = NUM2INT(rb_root);
   Data_Get_Struct(self, struct _Comm, comm);
   check_error(MPI_Comm_rank(comm->Comm, &rank));
   check_error(MPI_Comm_size(comm->Comm, &size));
   if (rank == root) {
-    OBJ2C(rb_recvbuf, recvcount, recvbuf, recvtype);
+    OBJ2C(rb_recvbuf, recvcount, recvbuf, recvtype, 0);
     if (recvcount != sendcount)
       rb_raise(rb_eArgError, "sendbuf and recvbuf has the same length");
     if (recvtype != sendtype)
@@ -502,16 +566,16 @@ static VALUE
 rb_comm_allreduce(VALUE self, VALUE rb_sendbuf, VALUE rb_recvbuf, VALUE rb_op)
 {
   void *sendbuf, *recvbuf;
-  int sendcount, recvcount;
+  int sendcount=0, recvcount=0;
   MPI_Datatype sendtype, recvtype;
   int rank, size;
   struct _Comm *comm;
   struct _Op *op;
-  OBJ2C(rb_sendbuf, sendcount, sendbuf, sendtype);
+  OBJ2C(rb_sendbuf, sendcount, sendbuf, sendtype, 0);
   Data_Get_Struct(self, struct _Comm, comm);
   check_error(MPI_Comm_rank(comm->Comm, &rank));
   check_error(MPI_Comm_size(comm->Comm, &size));
-  OBJ2C(rb_recvbuf, recvcount, recvbuf, recvtype);
+  OBJ2C(rb_recvbuf, recvcount, recvbuf, recvtype, 0);
   if (recvcount != sendcount)
     rb_raise(rb_eArgError, "sendbuf and recvbuf has the same length");
   if (recvtype != sendtype)
@@ -529,7 +593,8 @@ rb_comm_get_Errhandler(VALUE self)
 
   Data_Get_Struct(self, struct _Comm, comm);
   rb_errhandler = Data_Make_Struct(cErrhandler, struct _Errhandler, NULL, Errhandler_free, errhandler);
-  MPI_Comm_get_errhandler(comm->Comm, &(errhandler->Errhandler));
+  errhandler->free = false;
+  check_error(MPI_Comm_get_errhandler(comm->Comm, &(errhandler->Errhandler)));
   return rb_errhandler;
 }
 static VALUE
@@ -540,7 +605,7 @@ rb_comm_set_Errhandler(VALUE self, VALUE rb_errhandler)
 
   Data_Get_Struct(self, struct _Comm, comm);
   Data_Get_Struct(rb_errhandler, struct _Errhandler, errhandler);
-  MPI_Comm_set_errhandler(comm->Comm, errhandler->Errhandler);
+  check_error(MPI_Comm_set_errhandler(comm->Comm, errhandler->Errhandler));
   return self;
 }
 static VALUE
@@ -607,6 +672,7 @@ void Init_mpi()
   mMPI = rb_define_module("MPI");
   rb_define_module_function(mMPI, "Init", rb_m_init, -1);
   rb_define_module_function(mMPI, "Finalize", rb_m_finalize, -1);
+  rb_define_module_function(mMPI, "Abort", rb_m_abort, 2);
   rb_define_const(mMPI, "VERSION", INT2NUM(MPI_VERSION));
   rb_define_const(mMPI, "SUBVERSION", INT2NUM(MPI_SUBVERSION));
   rb_define_const(mMPI, "SUCCESS", INT2NUM(MPI_SUCCESS));
@@ -620,8 +686,8 @@ void Init_mpi()
   rb_define_method(cComm, "size", rb_comm_size, 0);
   rb_define_method(cComm, "Send", rb_comm_send, 3);
   rb_define_method(cComm, "Isend", rb_comm_isend, 3);
-  rb_define_method(cComm, "Recv", rb_comm_recv, 3);
-  rb_define_method(cComm, "Irecv", rb_comm_irecv, 3);
+  rb_define_method(cComm, "Recv", rb_comm_recv, -1);
+  rb_define_method(cComm, "Irecv", rb_comm_irecv, -1);
   rb_define_method(cComm, "Gather", rb_comm_gather, 3);
   rb_define_method(cComm, "Allgather", rb_comm_allgather, 2);
   rb_define_method(cComm, "Bcast", rb_comm_bcast, 2);
